#include "include/nsd_windows/nsd_windows_plugin.h"

#include <winsock2.h>
#pragma comment(lib,"Ws2_32.lib") 

// This must be included before many other Windows headers.
#include <windows.h>

// For getPlatformVersion; remove unless needed for your plugin implementation.
#include <VersionHelpers.h>

#include <flutter/method_channel.h>
#include <flutter/plugin_registrar_windows.h>
#include <flutter/standard_method_codec.h>

#include <map>
#include <memory>
#include <sstream>

#include <windns.h>
#pragma comment(lib,"dnsapi.lib")

namespace {
  // Converts the given UTF-8 string to UTF-16.
  std::wstring Utf16FromUtf8(const std::string& utf8_string) {
    if (utf8_string.empty()) {
      return std::wstring();
    }
    int target_length =
      ::MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, utf8_string.data(),
        static_cast<int>(utf8_string.length()), nullptr, 0);
    if (target_length == 0) {
      return std::wstring();
    }
    std::wstring utf16_string;
    utf16_string.resize(target_length);
    int converted_length =
      ::MultiByteToWideChar(CP_UTF8, MB_ERR_INVALID_CHARS, utf8_string.data(),
        static_cast<int>(utf8_string.length()),
        utf16_string.data(), target_length);
    if (converted_length == 0) {
      return std::wstring();
    }
    return utf16_string;
  }

class NsdWindowsPlugin : public flutter::Plugin {
 public:
  static void RegisterWithRegistrar(flutter::PluginRegistrarWindows *registrar);
  static std::unique_ptr<flutter::MethodChannel<flutter::EncodableValue>> methodChannel;

  NsdWindowsPlugin();

  virtual ~NsdWindowsPlugin();

protected:
  static void browseCallback(const DWORD status, void *context, DNS_RECORD *records);
	static void resolveCompletionCallback(const DWORD status, void *context, DNS_SERVICE_INSTANCE *instance);
  bool valid;

 private:
  // Called when a method is called on this plugin's channel from Dart.
  void HandleMethodCall(
      const flutter::MethodCall<flutter::EncodableValue> &method_call,
      std::unique_ptr<flutter::MethodResult<flutter::EncodableValue>> result);
  void startRegister(
      const flutter::MethodCall<flutter::EncodableValue> &method_call,
      std::unique_ptr<flutter::MethodResult<flutter::EncodableValue>> result);
  DWORD unregister(const flutter::MethodCall<flutter::EncodableValue> &method_call);
  void startBrowser(const flutter::MethodCall<flutter::EncodableValue> &method_call,
    std::unique_ptr<flutter::MethodResult<flutter::EncodableValue>> result);
  DWORD stopBrowser(const flutter::MethodCall<flutter::EncodableValue> &method_callt);
	void startResolver(const flutter::MethodCall<flutter::EncodableValue> &method_call,
    std::unique_ptr<flutter::MethodResult<flutter::EncodableValue>> result);
  DWORD stopResolver(const flutter::MethodCall<flutter::EncodableValue> &method_call);
  std::map<std::string, std::unique_ptr<DNS_SERVICE_CANCEL>> cancelRegisterMap = {};
  std::map<std::string, std::unique_ptr<DNS_SERVICE_CANCEL>> cancelBrowserMap = {};
  std::map<std::string, std::unique_ptr<DNS_SERVICE_CANCEL>> cancelResolverMap = {};
};

// static
void NsdWindowsPlugin::RegisterWithRegistrar(
    flutter::PluginRegistrarWindows *registrar) {
  methodChannel =
      std::make_unique<flutter::MethodChannel<flutter::EncodableValue>>(
          registrar->messenger(), "com.haberey/nsd",
          &flutter::StandardMethodCodec::GetInstance());

  auto plugin = std::make_unique<NsdWindowsPlugin>();

  methodChannel->SetMethodCallHandler(
      [plugin_pointer = plugin.get()](const auto &call, auto result) {
        plugin_pointer->HandleMethodCall(call, std::move(result));
      });
  registrar->AddPlugin(std::move(plugin));
}

std::unique_ptr<flutter::MethodChannel<flutter::EncodableValue>> NsdWindowsPlugin::methodChannel;

void callbackRegisterComplete(const DWORD status, void* context, DNS_SERVICE_INSTANCE* instance) {
  if (instance) {
  }

  if (status == ERROR_CANCELLED) {
    return;
  }

  if (status != ERROR_SUCCESS) {
  }
}
void NsdWindowsPlugin::browseCallback(const DWORD status, void* context, DNS_RECORD* records) {
  if (status == ERROR_CANCELLED) {
    return;
  }
  if (status != ERROR_SUCCESS) {
    return;
  }
  std::string* handle = static_cast<std::string*>(context);
  for (auto cur = records; cur; cur = cur->pNext) {
    if (cur->wType != DNS_TYPE_PTR) {
      continue;
    }

    // "_pasteshare._tcp.local".
    const wchar_t* pName = cur->pName;
    std::wstring domainW(pName);
    std::string domain(domainW.begin(), domainW.end());
    // "Test._pasteshare._tcp.local".
    wchar_t * pNameHost = cur->Data.PTR.pNameHost;
    std::wstring hostnameW(pNameHost);
    std::string hostname(hostnameW.begin(), hostnameW.end());
    
    std::string serviceName = hostname.substr(0, hostname.find(domain));
    // Trim trailing ".". 
    serviceName.resize(serviceName.size() - 1);

    std::string serviceType = domain.substr(0, domain.rfind('.'));

    NsdWindowsPlugin::methodChannel->InvokeMethod("onServiceDiscovered", std::make_unique<flutter::EncodableValue>(
      flutter::EncodableMap{
        {flutter::EncodableValue("handle"), flutter::EncodableValue(*handle)},
        {flutter::EncodableValue("service.type"), flutter::EncodableValue(serviceType)},
        {flutter::EncodableValue("service.name"), flutter::EncodableValue(serviceName)},
        {flutter::EncodableValue("service.host"), flutter::EncodableValue(hostname)}, 
      }));
  }
}
void NsdWindowsPlugin::resolveCompletionCallback(const DWORD status, void* context, DNS_SERVICE_INSTANCE* instance) {
  if (status == ERROR_CANCELLED) {
    return;
  }
  std::string* handle = static_cast<std::string*>(context);
  if (status != ERROR_SUCCESS) {
    NsdWindowsPlugin::methodChannel->InvokeMethod("onResolveFailed", std::make_unique<flutter::EncodableValue>(
    flutter::EncodableMap{
      {flutter::EncodableValue("handle"), flutter::EncodableValue(*handle)},
      {flutter::EncodableValue("error.cause"), flutter::EncodableValue("code:" + status)},
      {flutter::EncodableValue("error.message"), flutter::EncodableValue("Refer: https://docs.microsoft.com/en-us/windows/win32/debug/system-error-codes--0-499-")},
    }));
    return;
  }
  flutter::EncodableMap txtMap;
  for (int i = 0; i < instance->dwPropertyCount; i++) {
    wchar_t* keyA = instance->keys[i];
    std::wstring keyW(keyA);
    std::string key(keyW.begin(), keyW.end());
    wchar_t* valueA = instance->values[i];
    std::wstring valueAW(valueA);
    std::string valueString(valueAW.begin(), valueAW.end());
    std::vector<unsigned char> aTXT;
    for (int i = 0; i < wcslen(valueA); i++) {
        aTXT.push_back(valueString[i]);
    }
    txtMap.insert_or_assign(flutter::EncodableValue(key), flutter::EncodableValue(aTXT));
  }
  wchar_t * pszHostName = instance->pszHostName;
  std::wstring hostnameW(pszHostName);
  std::string hostname(hostnameW.begin(), hostnameW.end());
  wchar_t* pszInstanceName = instance->pszInstanceName;
  std::wstring instanceNameW(pszInstanceName);
  std::string instanceName(instanceNameW.begin(), instanceNameW.end());
  std::string local = hostname.substr(hostname.find("."));
  std::string name = instanceName.substr(0, instanceName.find("."));
  std::string serviceType = instanceName.substr(name.length()+1, instanceName.length() - local.length() - name.length() - 1);
  //DWORD* addressV4D = instance->ip4Address;
  //struct in_addr addr;
  //addr.S_un.S_addr = (long)addressV4D;
  //char* addressV4 = inet_ntoa(addr);
  //std::string ipV4Address = addressV4;
  NsdWindowsPlugin::methodChannel->InvokeMethod("onResolveSuccessful", std::make_unique<flutter::EncodableValue>(
    flutter::EncodableMap{
      {flutter::EncodableValue("handle"), flutter::EncodableValue(*handle)},
      {flutter::EncodableValue("service.type"), flutter::EncodableValue(serviceType)},
      {flutter::EncodableValue("service.name"), flutter::EncodableValue(name)},
      {flutter::EncodableValue("service.host"), flutter::EncodableValue(hostname)},
      {flutter::EncodableValue("service.txt"), txtMap},
    }));
}

NsdWindowsPlugin::NsdWindowsPlugin(): valid(false) {
  auto handle = GetModuleHandle(L"dnsapi.dll");
	if (handle) {
		if (DnsServiceConstructInstance && DnsServiceFreeInstance && DnsServiceRegister && DnsServiceDeRegister
			&& DnsServiceRegisterCancel) {
			valid = true;
		}
	}
}

NsdWindowsPlugin::~NsdWindowsPlugin() {}

template <class T>
T GetArgument(const flutter::MethodCall<>& method_call, const std::string key = "handle") {
  T value;
  const auto* arguments = std::get_if<flutter::EncodableMap>(method_call.arguments());
  if (arguments) {
    auto it = arguments->find(flutter::EncodableValue(key));
    if (it != arguments->end()) {
      value = std::get<T>(it->second);
    }
  }
  return value;
}

void NsdWindowsPlugin::HandleMethodCall(
    const flutter::MethodCall<flutter::EncodableValue> &method_call,
    std::unique_ptr<flutter::MethodResult<flutter::EncodableValue>> result) {
  std::string methodName = method_call.method_name();
  if (!valid) {
    result->Error("-1", "Windows Native API not available. Refer: https://docs.microsoft.com/en-us/windows/win32/api/windns/");
    return;
  }
  if (methodName.compare("startDiscovery") == 0) {
    NsdWindowsPlugin::stopBrowser(method_call);
    NsdWindowsPlugin::startBrowser(method_call, std::move(result));
  } else if (methodName.compare("stopDiscovery") == 0) {
    auto ret = NsdWindowsPlugin::stopBrowser(method_call);
    if (ret == ERROR_SUCCESS || ret == ERROR_CANCELLED) {
      result->Success(NULL);
    }
    result->Error(std::to_string(ret), "");
  } else if (methodName.compare("register") == 0) {
    NsdWindowsPlugin::unregister(method_call);
    NsdWindowsPlugin::startRegister(method_call, std::move(result));
  } else if (methodName.compare("resolve") == 0) {
    NsdWindowsPlugin::stopResolver(method_call);
    NsdWindowsPlugin::startResolver(method_call, std::move(result));
  } else if (methodName.compare("unregister") == 0) {
    auto ret = NsdWindowsPlugin::unregister(method_call);
    if (ret == ERROR_SUCCESS || ret == ERROR_CANCELLED) {
      result->Success(NULL);
    }
    result->Error(std::to_string(ret), "");
  } else {
    result->NotImplemented();
  }
}

void NsdWindowsPlugin::startRegister(const flutter::MethodCall<flutter::EncodableValue> &method_call,
  std::unique_ptr<flutter::MethodResult<flutter::EncodableValue>> result) {
    DWORD size = 0;
    GetComputerNameEx(ComputerNameDnsHostname, nullptr, &size);
    std::vector< wchar_t > hostname(size);
    if (!GetComputerNameEx(ComputerNameDnsHostname, &hostname[0], &size)) {
      result->Success(flutter::EncodableValue(NULL));
      return;
    }
    const std::wstring domain = L".local";
    const auto qualifiedHostname = &hostname[0] + domain;

    std::wstring service = &hostname[0];
    service += L".";
    std::string serviceType = GetArgument<std::string>(method_call, "service.type");
    service += Utf16FromUtf8(serviceType);
    service += domain;

    flutter::EncodableMap txtMap = GetArgument<flutter::EncodableMap>(method_call, "service.txt");
    int dwPropertiesCount = txtMap.size();
    auto it = txtMap.begin();
    std::vector<PCWSTR> keys;
    std::vector<PCWSTR> values;
    while (it != txtMap.end())
    {
      auto keyString = std::get<std::string>(it->first);
      auto keyWString = Utf16FromUtf8(keyString);
      const wchar_t* keyCString = keyWString.c_str();
      auto mallocCString = (wchar_t*)malloc(sizeof(wchar_t) * wcslen(keyCString));
      wcscpy(mallocCString, keyCString);
      keys.push_back(mallocCString);

      auto pValueChar = std::get<std::vector<unsigned char>>(it->second);
      std::string valeString(pValueChar.begin(), pValueChar.end());
      auto valueWString = Utf16FromUtf8(valeString);
      auto valueWCString = valueWString.c_str();
      auto mallocValueWCString = (wchar_t*)malloc(sizeof(wchar_t) * wcslen(valueWCString));
      wcscpy(mallocValueWCString, valueWCString);
      values.push_back(mallocValueWCString);
      it++;
    }
    auto instance = DnsServiceConstructInstance(service.c_str(), qualifiedHostname.c_str(), nullptr, nullptr, GetArgument<int>(method_call, "service.port"), 0, // 
      0, dwPropertiesCount, &keys[0], &values[0]);

    std::unique_ptr<DNS_SERVICE_REGISTER_REQUEST> m_request;
    m_request.reset(new DNS_SERVICE_REGISTER_REQUEST{});
    m_request->Version = DNS_QUERY_REQUEST_VERSION1;
    m_request->pServiceInstance = instance;
    m_request->pRegisterCompletionCallback = callbackRegisterComplete;
    m_request->pQueryContext = this;

    std::unique_ptr<DNS_SERVICE_CANCEL> registerCancel;
    registerCancel.reset(new DNS_SERVICE_CANCEL{});
    const auto ret = DnsServiceRegister(m_request.get(), registerCancel.get());
    DnsServiceFreeInstance(instance);

    result->Success(flutter::EncodableValue(NULL));
    std::string handle = GetArgument<std::string>(method_call, "handle");
    if (ret == DNS_REQUEST_PENDING) {
      cancelRegisterMap.insert_or_assign(handle, std::move(registerCancel));
      NsdWindowsPlugin::methodChannel->InvokeMethod("onRegistrationSuccessful", std::make_unique<flutter::EncodableValue>(
        flutter::EncodableMap{
          {flutter::EncodableValue("handle"), flutter::EncodableValue(handle)},
          {flutter::EncodableValue("service.type"), flutter::EncodableValue(serviceType)}
        }
      ));
      return;
    }
    NsdWindowsPlugin::methodChannel->InvokeMethod("onRegistrationFailed", std::make_unique<flutter::EncodableValue>(
        flutter::EncodableMap{
          {flutter::EncodableValue("handle"), flutter::EncodableValue(handle)},
          {flutter::EncodableValue("error.cause"), flutter::EncodableValue("code:" + ret)},
          {flutter::EncodableValue("error.message"), flutter::EncodableValue("Refer: https://docs.microsoft.com/en-us/windows/win32/debug/system-error-codes--0-499-")},
        }
      ));
  }

  DWORD NsdWindowsPlugin::unregister(const flutter::MethodCall<flutter::EncodableValue>& method_call) {
    std::string handle = GetArgument<std::string>(method_call, "handle");
    auto it = cancelRegisterMap.find(handle);
    if (it != cancelRegisterMap.end()) {
      return DnsServiceRegisterCancel(it->second.get());
    }
    return ERROR_SUCCESS;
  }
  
  void NsdWindowsPlugin::startBrowser(const flutter::MethodCall<flutter::EncodableValue> &method_call,
    std::unique_ptr<flutter::MethodResult<flutter::EncodableValue>> result) {
    std::string handle = GetArgument<std::string>(method_call, "handle");
    auto queryName = GetArgument<std::string>(method_call, "service.type") + ".local";
    DNS_SERVICE_BROWSE_REQUEST request{};
    request.Version = DNS_QUERY_REQUEST_VERSION1;
    auto queryNameW = Utf16FromUtf8(queryName);
    request.QueryName = reinterpret_cast<LPCWSTR>(queryNameW.c_str());
    request.pBrowseCallback = &NsdWindowsPlugin::browseCallback;
    auto sharedHandle = new std::string(handle);
    request.pQueryContext = sharedHandle;

    std::unique_ptr<DNS_SERVICE_CANCEL> cancelBrowser;
    cancelBrowser.reset(new DNS_SERVICE_CANCEL{});
    const auto ret = DnsServiceBrowse(&request, cancelBrowser.get());
    result->Success(flutter::EncodableValue(NULL));
    if (ret == DNS_REQUEST_PENDING) {
      cancelBrowserMap.insert_or_assign(handle, std::move(cancelBrowser));
      NsdWindowsPlugin::methodChannel->InvokeMethod("onDiscoveryStartSuccessful", std::make_unique<flutter::EncodableValue>(
        flutter::EncodableMap{
          {flutter::EncodableValue("handle"), flutter::EncodableValue(handle)},
        }));
      return;
    }
    NsdWindowsPlugin::methodChannel->InvokeMethod("onDiscoveryStartFailed", std::make_unique<flutter::EncodableValue>(
        flutter::EncodableMap{
          {flutter::EncodableValue("handle"), flutter::EncodableValue(handle)},
          {flutter::EncodableValue("error.cause"), flutter::EncodableValue("code:" + ret)},
          {flutter::EncodableValue("error.message"), flutter::EncodableValue("Refer: https://docs.microsoft.com/en-us/windows/win32/debug/system-error-codes--0-499-")},
        }
      ));
  }

  DWORD NsdWindowsPlugin::stopBrowser(const flutter::MethodCall<flutter::EncodableValue>& method_call) {
    std::string handle = GetArgument<std::string>(method_call, "handle");
    auto it = cancelBrowserMap.find(handle);
    if (it != cancelBrowserMap.end()) {
      return DnsServiceBrowseCancel(it->second.get());
    }
    return ERROR_SUCCESS;
  }

  void NsdWindowsPlugin::startResolver(const flutter::MethodCall<flutter::EncodableValue> &method_call,
    std::unique_ptr<flutter::MethodResult<flutter::EncodableValue>> result) {
    std::string handle = GetArgument<std::string>(method_call, "handle");
    auto qualifiedHostname = GetArgument<std::string>(method_call, "service.host");

    auto context = new std::string(handle);
    DNS_SERVICE_CANCEL cancle{
      context
    };
    DNS_SERVICE_RESOLVE_REQUEST request{};
    request.Version = DNS_QUERY_REQUEST_VERSION1;
    std::wstring qualifiedHostnameW = Utf16FromUtf8(qualifiedHostname);
    request.QueryName = &qualifiedHostnameW[0];
    request.pResolveCompletionCallback = &NsdWindowsPlugin::resolveCompletionCallback;
    request.pQueryContext = context;

    const auto ret = DnsServiceResolve(&request, &cancle);
    if (ret == DNS_REQUEST_PENDING) {
    }
    result->Success(flutter::EncodableValue(NULL));
  }

  DWORD NsdWindowsPlugin::stopResolver(const flutter::MethodCall<flutter::EncodableValue>& method_call) {
    std::string handle = GetArgument<std::string>(method_call, "handle");
    auto it = cancelResolverMap.find(handle);
    if (it != cancelResolverMap.end()) {
      return DnsServiceResolveCancel(it->second.get());
    }
    return ERROR_SUCCESS;
  }

}  // namespace

void NsdWindowsPluginRegisterWithRegistrar(
    FlutterDesktopPluginRegistrarRef registrar) {
  NsdWindowsPlugin::RegisterWithRegistrar(
      flutter::PluginRegistrarManager::GetInstance()
          ->GetRegistrar<flutter::PluginRegistrarWindows>(registrar));
}
